# Copyright [2024] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from dataclasses import dataclass
from ...math.model import TokenDeltaModel

DEFAULT_CHAIN_NM = "api.0x.org"
DEFAULT_BUY_TKN_NM = "WETH"
DEFAULT_BUY_TOKEN = '0xc02aaa39b223fe8d0a0e5c4f27ead9083c756cc2'
DEFAULT_SELL_TKN_NM = "USDC"
DEFAULT_SELL_TOKEN = '0xa0b86991c6218b36c1d19d4a2e9eb0ce3606eb48'
DEFAULT_INIT_AMT = 1000
DEFAULT_TIME_WINDOW = 0.25
DEFAULT_TRADE_BIAS = 0.5
DEFAULT_INIT_INVESTMENT = 1
DEFAULT_GAMMA_SHAPE = 1
DEFAULT_GAMMA_SCALE = 1
DEFAULT_MAX_TRADE_PERCENT = 0.003
DEFAULT_API_KEY = "6cbf2275-5cee-4659-8d67-5491399a9c5e"
DEFAULT_API_SELL_AMOUNT = '10000000'

@dataclass
class Chain0x:
    
    """ 
        Data class handling all data for BUIDL week ETHDenver Simulator application
    """    
    
    # API Endpoints for Chains
    ETHEREUM = 'api.0x.org'
    ARBITRUM = 'arbitrum.api.0x.org'
    AVALANCHE = 'avalanche.api.0x.org'
    BASE = 'base.api.0x.org'
    BINANCE = 'bsc.api.0x.org'
    CELO = 'celo.api.0x.org'
    FANTOM = 'fantom.api.0x.org'
    OPTIMISM = 'optimism.api.0x.org'
    POLYGON = 'polygon.api.0x.org'    
    
    # Buy tokens
    WETH = "WETH"
    LINK = "LINK"
    UNI = "UNI"
    WBTC = "WBTC"
    BNB = "BNB"  
    
    # Sell tokens
    USDC = "USDC"
    USDT = "USDT"
    DAI = "DAI"
    
    # Set defaults
    chain_nm: str = DEFAULT_CHAIN_NM
    buy_tkn_nm: str = DEFAULT_BUY_TKN_NM
    sell_tkn_nm: str = DEFAULT_SELL_TKN_NM
    max_trade_percent: float = DEFAULT_MAX_TRADE_PERCENT
    time_window: float = DEFAULT_TIME_WINDOW
    trade_bias: float = DEFAULT_TRADE_BIAS 
    init_investment: float = DEFAULT_INIT_INVESTMENT 
    api_key: float = DEFAULT_API_KEY
    api_sell_amount: str = DEFAULT_API_SELL_AMOUNT
    
    def get_api_key(self) -> str:
        
        """ get_api_key

            Get 0x API key
                
            Returns
            -----------------
            api_key : str
                0x API key                 
        """          
        
        return self.api_key
    
    def get_api_sell_amount(self) -> str:
        
        """ get_api_sell_amount

            Get 0x API sell amount setting
                
            Returns
            -----------------
            api_sell_amount : str
                0x API sell amount setting                
        """          
        
        return self.api_sell_amount    

    def get_chain_name(self) -> str:
        
        """ get_chain_name

            Get URL name for 0x api
                
            Returns
            -----------------
            select_chain_name : str
                URL name for 0x api             
        """            

        match self.chain_nm:
            case self.ETHEREUM:
                select_chain_name = 'api.0x.org'
            case self.ARBITRUM:
                select_chain_name = 'arbitrum.api.0x.org'
            case self.AVALANCHE:
                select_chain_name = 'avalanche.api.0x.org'
            case self.BASE:
                select_chain_name = 'base.api.0x.org'
            case self.BINANCE:
                select_chain_name = 'bsc.api.0x.org'
            case self.CELO:
                select_chain_name = 'celo.api.0x.org'
            case self.FANTOM:
                select_chain_name = 'fantom.api.0x.org'
            case self.OPTIMISM:
                select_chain_name = 'optimism.api.0x.org'
            case self.POLYGON:
                select_chain_name = 'polygon.api.0x.org'    

        return select_chain_name   
    
    def get_buy_token(self) -> str:
        
        """ get_buy_token

            Get buy token contract address for 0x api
                
            Returns
            -----------------
            select_buy_token : str
                Buy token contract address for 0x api           
        """         

        match self.buy_tkn_nm:
            case self.WETH if self.chain_nm in self.ETHEREUM:
                select_buy_token = '0xc02aaa39b223fe8d0a0e5c4f27ead9083c756cc2'
            case self.LINK if self.chain_nm in self.ETHEREUM:
                select_buy_token = '0x514910771AF9Ca656af840dff83E8264EcF986CA'
            case self.UNI if self.chain_nm in self.ETHEREUM:
                select_buy_token = '0x1f9840a85d5aF5bf1D1762F925BDADdC4201F984'
            case self.WBTC if self.chain_nm in self.ETHEREUM:
                select_buy_token = '0x2260FAC5E5542a773Aa44fBCfeDf7C193bc2C599'
            case self.BNB if self.chain_nm in self.ETHEREUM:
                select_buy_token = '0xB8c77482e45F1F44dE1745F52C74426C631bDD52'
                
            case self.WETH if self.chain_nm in self.POLYGON:
                select_buy_token = '0x7ceB23fD6bC0adD59E62ac25578270cFf1b9f619'
            case self.LINK if self.chain_nm in self.POLYGON:
                select_buy_token = '0xb0897686c545045aFc77CF20eC7A532E3120E0F1'
            case self.UNI if self.chain_nm in self.POLYGON:
                select_buy_token = '0xb33EaAd8d922B1083446DC23f610c2567fB5180f'
            case self.WBTC if self.chain_nm in self.POLYGON:
                select_buy_token = '0x1BFD67037B42Cf73acF2047067bd4F2C47D9BfD6'
            case self.BNB if self.chain_nm in self.POLYGON:
                select_buy_token = '0x3BA4c387f786bFEE076A58914F5Bd38d668B42c3'  
                
            case self.WETH if self.chain_nm in self.AVALANCHE:
                select_buy_token = '0x49D5c2BdFfac6CE2BFdB6640F4F80f226bc10bAB'
            case self.LINK if self.chain_nm in self.AVALANCHE:
                select_buy_token = '0x5947BB275c521040051D82396192181b413227A3'
            case self.UNI if self.chain_nm in self.AVALANCHE:
                select_buy_token = '0x8eBAf22B6F053dFFeaf46f4Dd9eFA95D89ba8580'
            case self.WBTC if self.chain_nm in self.AVALANCHE:
                select_buy_token = '0x50b7545627a5162F82A992c33b87aDc75187B218'
            case self.BNB if self.chain_nm in self.AVALANCHE:
                select_buy_token = '0x264c1383EA520f73dd837F915ef3a732e204a493'                   
                
            case self.WETH if self.chain_nm in self.OPTIMISM:
                select_buy_token = '0xe50fA9b3c56FfB159cB0FCA61F5c9D750e8128c8'
            case self.LINK if self.chain_nm in self.OPTIMISM:
                select_buy_token = '0x350a791Bfc2C21F9Ed5d10980Dad2e2638ffa7f6'
            case self.WBTC if self.chain_nm in self.OPTIMISM:
                select_buy_token = '0x68f180fcCe6836688e9084f035309E29Bf0A2095'                 

        return select_buy_token   
    
    def get_sell_token(self) -> str:
        
        """ get_buy_token

            Get sell token contract address for 0x api
                
            Returns
            -----------------
            select_buy_token : str
                Sell token contract address for 0x api           
        """           
        
        match self.sell_tkn_nm:
            case self.USDC if self.chain_nm in self.ETHEREUM:
                select_sell_token = '0xa0b86991c6218b36c1d19d4a2e9eb0ce3606eb48'
            case self.USDT if self.chain_nm in self.ETHEREUM:
                select_sell_token = '0xdAC17F958D2ee523a2206206994597C13D831ec7'
            case self.DAI if self.chain_nm in self.ETHEREUM:
                select_sell_token = '0x6B175474E89094C44Da98b954EedeAC495271d0F'
                
            case self.USDC if self.chain_nm in self.POLYGON:
                select_sell_token = '0x3c499c542cEF5E3811e1192ce70d8cC03d5c3359'
            case self.USDT if self.chain_nm in self.POLYGON:
                select_sell_token = '0xc2132D05D31c914a87C6611C10748AEb04B58e8F'
            case self.DAI if self.chain_nm in self.POLYGON:
                select_sell_token = '0x8f3Cf7ad23Cd3CaDbD9735AFf958023239c6A063'    
                
            case self.USDC if self.chain_nm in self.OPTIMISM:
                select_sell_token = '0x0b2C639c533813f4Aa9D7837CAf62653d097Ff85'
            case self.USDT if self.chain_nm in self.OPTIMISM:
                select_sell_token = '0x9702230A8Ea53601f5cD2dc00fDBc13d4dF4A8c7'
            case self.DAI if self.chain_nm in self.OPTIMISM:
                select_sell_token = '0x94b008aA00579c1307B0EF2c499aD98a8ce58e58'      
                
            case self.USDC if self.chain_nm in self.AVALANCHE:
                select_sell_token = '0xB97EF9Ef8734C71904D8002F8b6Bc66Dd9c48a6E'
            case self.USDT if self.chain_nm in self.AVALANCHE:
                select_sell_token = '0x9702230A8Ea53601f5cD2dc00fDBc13d4dF4A8c7'
            case self.DAI if self.chain_nm in self.AVALANCHE:
                select_sell_token = '0xbA7dEebBFC5fA1100Fb055a87773e1E99Cd3507a'                  

                
        return select_sell_token      
    
    def get_buy_init_amt(self) -> float:
        
        """ get_buy_init_amt

            Get x init amount for pool initialization
                
            Returns
            -----------------
            select_buy_init_amt : int
                x init amount for pool initialization         
        """           

        match self.buy_tkn_nm:
            case self.WETH:
                select_buy_init_amt = 10000
            case self.LINK:
                select_buy_init_amt = 1000000
            case self.UNI:
                select_buy_init_amt = 1000000
            case self.WBTC:
                select_buy_init_amt = 1000
            case self.BNB:
                select_buy_init_amt = 100000

        return select_buy_init_amt 
    
    def get_td_model(self) -> TokenDeltaModel:
        
        """ get_td_model

            Get token delta model which is the non-deterministic model for incoming swap amounts; set 
            to Gamma distribution with paramaters set to scale = 1 and shape = 1 (by default)
                
            Returns
            -----------------
            td_model : TokenDeltaModel
                Token delta model      
        """             
        
        buy_init_amt = self.get_buy_init_amt()
        return TokenDeltaModel(max_trade = self.max_trade_percent*buy_init_amt, 
                                    shape = DEFAULT_GAMMA_SHAPE, # Gamma Dist. shape  
                                    scale = DEFAULT_GAMMA_SCALE) # Gamma Dist. scale     
    
    
             
                
                
                



# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from ..Process import Process
from ..swap import Swap
from ...cpt.vault import IndexVault
from ...math.model import TokenDeltaModel
from ...math.model import EventSelectionModel

class IndexTokenBurn(Process):

    """ Index token burn process

        Parameters
        ----------
        ivault : IndexVault
            Index vault
    """       
    
    def __init__(self, ivault):
        self.ivault = IndexVault('iVault', "0x7") if ivault  == None else ivault
            
    def apply(self, lp_tkn, token, user_nm, lp_burn_amt): 
        
        """ apply

            Burn index token from vault and swap out its parent token from its respective LP
                
            Parameters
            -------
            lp_tkn : Exchange
                LP exchange
            token : ERC20
                specified ERC20 token               
            user_nm : str
                account name
            lp_burn_amt : float
               token amount to be burned 
                
            Returns
            -------
            result : float
                amount of burned token                   
        """           
        
        tkn_amts = self.ivault.burn_lp_token(user_nm, lp_tkn, token, lp_burn_amt)
        result = -1

        if(bool(tkn_amts)):
            trading_token = self.get_trading_token(lp_tkn, token)
            out = Swap().apply(lp_tkn, trading_token, user_nm, tkn_amts[trading_token.token_name])
            result = out+tkn_amts[token.token_name]
            
        self.ivault.update_accounts(lp_tkn, token)    
        
        return result  
    
    def get_trading_token(self, lp, token):
        
        """ get_trading_token

            Get opposing token from specified token
                
            Parameters
            -------
            lp : Exchange
                LP exchange
            token : ERC20
                specified ERC20 token      
                
            Returns
            -------
            trading_token : ERC20 
                opposing ERC20 token                   
        """         
        
        tokens = lp.factory.exchange_to_tokens[lp.name]
        trading_token = tokens[lp.token1] if token.token_name == lp.token0 else tokens[lp.token0]
        return trading_token  
        



# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

import numpy as np
import pandas as pd
from termcolor import colored
from ...cpt.quote import LPQuote
from ...process.swap import Swap
from ...math.model import TokenDeltaModel

class MarkovState():
    
    def __init__(self, stochastic = True, tDel = None):
        self.tDel = TokenDeltaModel(500) if tDel == None else tDel
        self.stochastic = stochastic
        self.current_state = [[0, 0, 0, 0]]
        self.states = self.current_state

    def gen_states(self, N):   
        for k in range(N):
            self.next_state(self.tDel.delta())
        return self.get_state_df()          
        
    def next_state(self, minted = None):
        P = self.gen_trans_matrix(self.stochastic)     
        self.current_state = np.dot(self.current_state, P)
        minted = self.tDel.delta() if minted == None else minted
        self.current_state = self.current_state  + np.array([[minted, 0.0, 0.0, 0.0]])
        self.states = np.append(self.states, self.current_state, axis=0)
        
    def update_current_state(self, amt, stat_cat = 'Vault'):
        if(stat_cat == 'Vault'):
            # Update between held and vault
            self.states[-1,1] += amt  # Add to held
            self.states[-1,2] -= amt  # Remove from vault
             
    def get_current_state(self, colnm = None):  
        df_states = self.get_state_df(self.states[-2:])
        return df_states.iloc[-1] if colnm == None else df_states[colnm].iloc[-1]
    
    def get_state_df(self, state_arr = []): 
        state_arr = self.states[1:,] if len(state_arr) == 0 else state_arr
        df_states = pd.DataFrame(self.states[1:,])    
        df_states.columns = ['Mint', 'Held', 'Vault', 'Burned'] 
        df_states['dHeld'] = np.insert(np.diff(df_states['Held'].values), 0, 0)
        df_states['dVault'] = np.insert(np.diff(df_states['Vault'].values), 0, 0)
        df_states['dBurned'] = np.insert(np.diff(df_states['Burned'].values), 0, 0) 
        return df_states
        
    def scale_x(self, x):
        return list(x/np.sum(x))
    
    def rBeta(self, a, b, mx = None, mn =None):
        p = np.random.beta(a,b)
        p = p if mn == None else max(p, mn)
        p = p if mx == None else min(p, mx)                  
        return p   

    def gen_trans_matrix(self, stochastic = True):

        P_stochastic = np.array([ [0, 1, 0, 0],
                           [0, self.rBeta(19,81), self.rBeta(4,1), self.rBeta(1,99)],
                           #[0, self.rBeta(1,4, mn = 0.05, mx = 0.4), self.rBeta(4,1, mn = 0.6, mx = 0.95), 0],
                           [0, self.rBeta(1,4, mn = 0.2, mx = 0.3), self.rBeta(4,1, mn = 0.7, mx = 0.8), 0],
                           [0, 0, 0, 1]])

        P_deterministic = np.array([ [0, 1, 0, 0],
                           [0, 0.18, 0.8, 0.02],
                           [0, 0.2, 0.8, 0],
                           [0, 0, 0, 1]])    

        P = P_stochastic if stochastic else P_deterministic  
        return np.array(list(map(self.scale_x, P))) 
    
    def inspect_states(self, tail = True, num_states = 5):
        dfDistrLP1 = self.get_state_df()
        return dfDistrLP1.tail(num_states) if tail else dfDistrLP1.head(num_states)        
    
    def check_states(self):
        dfDistrLP = self.get_state_df()
        states_balanced = round(sum(dfDistrLP.iloc[-1,1:4]),8) == round(sum(dfDistrLP.iloc[:-1,0]),8)
        test_outcome = colored('PASS', 'green', attrs=['bold']) if states_balanced else colored('FAIL', 'red', attrs=['bold'])            
        print(f'Amount of tokens retained across states: {test_outcome}')       
        


# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from ..Process import Process
from ...math.model import TokenDeltaModel
from ...math.model import EventSelectionModel
from ...process.swap import Swap

class RandomSwap(Process):
    
    """ Process to randomly swap token X for token Y (and vice verse) 

        Parameters
        ----------
        ev : EventSelectionModel
            EventSelectionModel object to randomly generate buy vs sell events
        tDel : TokenDeltaModel
            TokenDeltaModel to randomly generate token amounts                 
    """           

    def __init__(self, p = None, ev = None, tDel = None):
        self.ev = EventSelectionModel() if ev  == None else ev
        self.tDel = TokenDeltaModel(50) if tDel == None else tDel
        self.buy_sell_prob = 0.5 if p == None else p
        
    def set_buy_sell_prob(self, p):
        self.buy_sell_prob = p
            
    def apply(self, lp, token_in, user_nm, amount_in = None):    
        
        """ apply

            Randomly swap token X for token Y (and vice verse) 
                
            Parameters
            -------
            lp : Exchange
                LP exchange
            token_in : ERC20
                specified ERC20 token               
            user_nm : str
                account name
            amount_in : float
               token amount to be swap 
                
            Returns
            -------
            amount_out_expected : float
                exchanged token amount               
        """         
        
        amount_in = self.tDel.delta() if amount_in == None else amount_in
        direction = self.ev.bi_select(self.buy_sell_prob) 
        tokens = lp.factory.token_from_exchange[lp.name]
        
        
        #print('direction {}, p {}'.format(direction, self.buy_sell_prob))

        if(direction == 0):
            token_in = tokens[lp.token0]
            amount_out_expected = Swap().apply(lp, token_in, user_nm, amount_in)
        elif(direction == 1):
            token_in = tokens[lp.token1]
            amount_out_expected = Swap().apply(lp, token_in, user_nm, amount_in)

        return amount_out_expected    
        



# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from ..Process import Process
from .Swap import Swap
from ..liquidity import RemoveLiquidity
from ...math.model import TokenDeltaModel
from ...math.model import EventSelectionModel
from ...utils.data import UniswapExchangeData
from ...cpt.index import SettlementLPToken
from ...utils.tools.v3 import FullMath
import math

class WithdrawSwap(Process):
    
    """ Process to withdraw liquidity from LP and swap opposing token which is added to specifed token to receive         
        a single amount of specified token

        Parameters
        ----------
        ev : EventSelectionModel
            EventSelectionModel object to randomly generate buy vs sell events
        tDel : TokenDeltaModel
            TokenDeltaModel to randomly generate token amounts        
    """       

    def __init__(self, ev = None, tDel = None):
        self.ev = EventSelectionModel() if ev  == None else ev
        self.tDel = TokenDeltaModel(50) if tDel == None else tDel
            
    def apply(self, lp, token_out, user_nm, amount_out, lwr_tick = None, upr_tick = None):    
        
        """ apply

            Withdraw liquidity from LP and swap opposing token which is added to specifed token to receive                 
            a single amount of specified token
                
            Parameters
            -------
            lp : Exchange
                LP exchange
            token_out : ERC20
                specified ERC20 token               
            user_nm : str
                account name
            amount_out : float
                token amount to be swap 
            lwr_tick : int
                lower tick of the position in which to add liquidity   
            upr_tick : int
                upper tick of the position in which to add liquidity                 
                
            Returns
            -------
            amount_out : float
                amount of withdrawn token               
        """          
        
        amount_out = self.tDel.delta() if amount_out == None else amount_out
      
        if(lp.version == UniswapExchangeData.VERSION_V2):            
            # Step 1: withdrawal
            p_out = self._calc_univ2_withdraw_portion(lp, token_out, amount_out, lwr_tick, upr_tick)
            removeLiq = RemoveLiquidity()   
            res = removeLiq.apply(lp, token_out, user_nm, p_out*amount_out)
    
            # Step 2: swap
            trading_token = self.get_trading_token(lp, token_out)
            out = Swap().apply(lp, trading_token, user_nm, res[trading_token.token_name])  
            withdrawn = out + p_out*amount_out 

        elif(lp.version == UniswapExchangeData.VERSION_V3): 

            p_out = self._calc_univ3_withdraw_portion(lp, token_out, amount_out, lwr_tick, upr_tick)
            
             # Step 1: withdrawal
            removeLiq = RemoveLiquidity()
            res = removeLiq.apply(lp, token_out, user_nm, p_out*amount_out, lwr_tick, upr_tick)
        
            # Step 2: swap
            trading_token = self.get_trading_token(lp, token_out)
            out = Swap().apply(lp, trading_token, user_nm, res[trading_token.token_name]) 
            withdrawn = abs(out)  + p_out*amount_out 

        return withdrawn 


    def get_trading_token(self, lp, token):
        
        """ get_trading_token

            Get opposing token from specified token
                
            Parameters
            -------
            lp : Exchange
                LP exchange
            token : ERC20
                specified ERC20 token      
                
            Returns
            -------
            trading_token : ERC20 
                opposing ERC20 token                   
        """          
        
        tokens = lp.factory.token_from_exchange[lp.name]
        trading_token = tokens[lp.token1] if token.token_name == lp.token0 else tokens[lp.token0]
        return trading_token  

    def _calc_univ3_withdraw_portion(self, lp, token_in, amt, lwr_tick, upr_tick):

        (x, y) = self._get_reserves(lp, token_in)
        L = lp.get_liquidity()
        gamma = 997/1000

        dL = SettlementLPToken().apply(lp, token_in, amt, lwr_tick, upr_tick)
        dx = dL*x/L
        dy = dL*y/L
        aswap = (gamma*dx)*(y-dy)/(x-dx+gamma*dx)

        return dy/amt 
 
    def _calc_univ2_withdraw_portion(self, lp, token_in, amt, lwr_tick, upr_tick):

        (x, y) = self._get_reserves(lp, token_in)

        L = lp.total_supply
        gamma = 997
        dL = SettlementLPToken().apply(lp, token_in, amt, lwr_tick, upr_tick)
        amt = lp.convert_to_machine(amt)
        dx = FullMath.divRoundingUp(dL*x,L)
        dy = FullMath.divRoundingUp(dL*y,L)
        aswap = (gamma*dx)*(y-dy)/(1000*x-1000*dx+gamma*dx)        

        return dy/amt 

    def _get_reserves(self, lp, token_in):
        tokens = lp.factory.token_from_exchange[lp.name]
        if(lp.version == UniswapExchangeData.VERSION_V2):
            if(token_in.token_name == lp.token1):
                x = lp.reserve0
                y = lp.reserve1
            else: 
                x = lp.reserve1
                y = lp.reserve0

        elif(lp.version == UniswapExchangeData.VERSION_V3):   
            if(token_in.token_name == lp.token1):
                x = lp.get_virtual_reserve(tokens[lp.token0])
                y = lp.get_virtual_reserve(tokens[lp.token1])
            else: 
                x = lp.get_virtual_reserve(tokens[lp.token1])
                y = lp.get_virtual_reserve(tokens[lp.token0]) 
        return (x, y)        
    
        



# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

import numpy as np
import math
from ...utils.data import UniswapExchangeData
from ...utils.tools.v3 import TickMath
from ...utils.tools.v3 import UniV3Helper
from ...utils.tools.v3 import FullMath

class SettlementLPToken():
    
    """ Determine settlement amount of LP token given a certain amount of token    
    """  
    
    def __init__(self):
        pass
     
    def apply(self, lp, tkn, itkn_amt, lwr_tick = None, upr_tick = None): 
        
        """ apply

            Apply LP calculation settlement; given token amount, what is the liquidity amount
                
            Parameters
            -----------------
            lp : UniswapExchange
                Uniswap LP    
            tkn: ERC20
                Token asset from CPT pair       
            itkn_amt: float
                Token reserve amount to be priced in terms of liquidity                 

            Returns
            -----------------
            lp_amount: float
                Liquidity amount given reserve token amount
                   
        """          
        if(lp.version == UniswapExchangeData.VERSION_V2):
            settlement = self.calc_univ2_lp_settlement(lp, tkn, itkn_amt)
        elif(lp.version == UniswapExchangeData.VERSION_V3):   
            settlement = self.calc_univ3_lp_settlement(lp, tkn, itkn_amt, lwr_tick, upr_tick)
        
        return settlement
    
    def calc_univ2_lp_settlement(self, lp, token_in, itkn_amt):

        (x, y) = self.get_reserves(lp, token_in)
        L = lp.get_liquidity()

        x = lp.convert_to_machine(x)
        y = lp.convert_to_machine(y)
        L = lp.convert_to_machine(L)
        itkn_amt = lp.convert_to_machine(itkn_amt)
        
        if(L == 0): 
            return 0
        else:
            gamma = 997
        
            a1 = FullMath.divRoundingUp(x*y,L)
            a2 = L
            a = FullMath.divRoundingUp(a1,a2)
            b = FullMath.divRoundingUp(1000*itkn_amt*x - itkn_amt*gamma*x + 1000*x*y + x*y*gamma,1000*L)
            c = itkn_amt*x;
            
            radicand = b*b - FullMath.divRoundingUp(4*a1*c,a2)
            dL = FullMath.divRoundingUp(b*a2 - a2*math.isqrt(radicand), 2*a1)
            return dL

    def calc_univ3_lp_settlement(self, lp, token_in, itkn_amt, lwr_tick, upr_tick):
    
        L = lp.get_liquidity()
        if(token_in.token_name == lp.token0):
            sqrtp_cur = lp.slot0.sqrtPriceX96/2**96
            sqrtp_pa = TickMath.getSqrtRatioAtTick(lwr_tick)/2**96
            sqrtp_pb = TickMath.getSqrtRatioAtTick(upr_tick)/2**96 
            dPy = (sqrtp_cur - sqrtp_pa)
            dPx = (1/sqrtp_cur - 1/sqrtp_pb)          
        elif(token_in.token_name == lp.token1):
            sqrtp_cur = 2**96/lp.slot0.sqrtPriceX96
            sqrtp_pa = 2**96/TickMath.getSqrtRatioAtTick(lwr_tick)
            sqrtp_pb = 2**96/TickMath.getSqrtRatioAtTick(upr_tick)
            dPx = (1/sqrtp_cur - 1/sqrtp_pa)
            dPy = (sqrtp_cur - sqrtp_pb)
    
        fee = 997
        
        a = fee*dPy*sqrtp_cur*dPx - 1000*dPx*(sqrtp_cur**2) - fee*dPy  
        b = -fee*dPy*sqrtp_cur*itkn_amt + 1000*itkn_amt*(sqrtp_cur**2) + L*fee*dPy + 1000*L*dPx*(sqrtp_cur**2)
        c = -1000*L*itkn_amt*(sqrtp_cur**2)
    
        return (-b + math.sqrt(b*b - 4*a*c)) / (2*a)  

    def get_reserves(self, lp, token_in):
        tokens = lp.factory.token_from_exchange[lp.name]

        if(token_in.token_name == lp.token1):
            x = lp.get_reserve(tokens[lp.token0])
            y = lp.get_reserve(tokens[lp.token1])
        else: 
            x = lp.get_reserve(tokens[lp.token1])
            y = lp.get_reserve(tokens[lp.token0])

        return (x, y)  
        
            


from uniswappy.erc import *
from uniswappy.cpt.exchg import *
from uniswappy.cpt.factory import *
from uniswappy.cpt.index import *
from uniswappy.cpt.quote import *
from uniswappy.cpt.vault import *
from uniswappy.cpt.wallet import *
from uniswappy.math.basic import *
from uniswappy.math.interest import *
from uniswappy.math.interest.ips import *
from uniswappy.math.interest.ips.aggregate import *
from uniswappy.math.model import *
from uniswappy.math.risk import *
from uniswappy.process import *
from uniswappy.process.burn import *
from uniswappy.process.deposit import *
from uniswappy.process.liquidity import *
from uniswappy.process.mint import *
from uniswappy.process.swap import *
from uniswappy.process.join import *
from uniswappy.analytics.simulate import *
from uniswappy.analytics.risk import *
from uniswappy.utils.interfaces import *
from uniswappy.utils.data import *
from uniswappy.utils.client import *
from uniswappy.utils.tools import *
from uniswappy.utils.tools.v3 import *


from setuptools import setup

with open('README.md') as f:
    long_description = f.read()

setup(name='UniswapPy',
      version='1.7.0',
      description='Uniswap Analytics with Python',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://github.com/defipy-devs/uniswappy',
      author = "icmoore",
      author_email = "defipy.devs@gmail.com",
      license='MIT',
      package_dir = {"uniswappy": "python/prod"},
      packages=[
          'uniswappy',
          'uniswappy.cpt.exchg',
          'uniswappy.cpt.factory',
          'uniswappy.cpt.index',
          'uniswappy.cpt.quote',
          'uniswappy.cpt.vault',
          'uniswappy.cpt.wallet',
          'uniswappy.erc',
          'uniswappy.math.basic',
          'uniswappy.math.interest',
          'uniswappy.math.interest.ips',
          'uniswappy.math.interest.ips.aggregate',
          'uniswappy.math.model',
          'uniswappy.math.risk',
          'uniswappy.process',
          'uniswappy.process.burn',
          'uniswappy.process.deposit',
          'uniswappy.process.liquidity',
          'uniswappy.process.mint',
          'uniswappy.process.swap',
          'uniswappy.process.join',
          'uniswappy.analytics.simulate',
          'uniswappy.analytics.risk',
          'uniswappy.utils.interfaces',
          'uniswappy.utils.data',
          'uniswappy.utils.client',
          'uniswappy.utils.tools',
          'uniswappy.utils.tools.v3'
      ],
      install_requires=[
          'scipy >= 1.7.3',
          'termcolor >= 2.4.0'
      ],
      zip_safe=False)



# Copyright [2024] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from dataclasses import dataclass

@dataclass
class LPType:
    
    HYBRID = 'HYBRID'
    SYNTHETIC = 'SYNTHETIC'



# Copyright [2023] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from ..basic import RoundFloat

class CompoundReturn():        
       
    __DEC = 5    
        
    def apply(self, A0, n_compounds, ips, t_ratio = 0.3):
        res = (A0)*(1+ips)**(n_compounds)
        return RoundFloat().apply(res, t_ratio)      


# Copyright [2024] [Ian Moore]
# Distributed under the MIT License (license terms are at http://opensource.org/licenses/MIT).
# Email: defipy.devs@gmail.com

from abc import *
from .IExchange import IExchange 

class IExchangeFactory(ABC):
    
    @abstractmethod        
    def deploy(self, agents):
        pass
    
    @abstractmethod        
    def get_exchange(self, exchange: IExchange):
        pass
    
    @abstractmethod        
    def get_token(self, exchange: IExchange):
        pass    